@@ -1,5 +1,5 @@
-/* stat.c -- display file or filesystem status
-   Copyright (C) 2001, 2002, 2003 Free Software Foundation.
+/* stat.c -- display file or file system status
+   Copyright (C) 2001, 2002, 2003, 2004, 2005 Free Software Foundation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
    Written by Michael Meskes.  */
 
@@ -49,8 +49,10 @@
 #include "file-type.h"
 #include "fs.h"
 #include "getopt.h"
+#include "inttostr.h"
 #include "quote.h"
 #include "quotearg.h"
+#include "stat-time.h"
 #include "strftime.h"
 #include "xreadlink.h"
 
@@ -58,18 +60,27 @@
 
 #if HAVE_STRUCT_STATVFS_F_BASETYPE
 # define STRUCT_STATVFS struct statvfs
-# define HAVE_STRUCT_STATXFS_F_TYPE HAVE_STRUCT_STATVFS
+# define HAVE_STRUCT_STATXFS_F_TYPE HAVE_STRUCT_STATVFS_F_TYPE
 # if HAVE_STRUCT_STATVFS_F_NAMEMAX
 #  define SB_F_NAMEMAX(S) ((uintmax_t) ((S)->f_namemax))
 # endif
+# if STAT_STATVFS
+#  define STATFS statvfs
+#  define STATFS_FRSIZE(S) ((S)->f_frsize)
+# endif
 #else
 # define STRUCT_STATVFS struct statfs
-# define HAVE_STRUCT_STATXFS_F_TYPE HAVE_STRUCT_STATFS
+# define HAVE_STRUCT_STATXFS_F_TYPE HAVE_STRUCT_STATFS_F_TYPE
 # if HAVE_STRUCT_STATFS_F_NAMELEN
 #  define SB_F_NAMEMAX(S) ((uintmax_t) ((S)->f_namelen))
 # endif
 #endif
 
+#ifndef STATFS
+# define STATFS statfs
+# define STATFS_FRSIZE(S) 0
+#endif
+
 #ifndef SB_F_NAMEMAX
 /* NetBSD 1.5.2 has neither f_namemax nor f_namelen.  */
 # define SB_F_NAMEMAX(S) "*"
@@ -90,33 +101,29 @@
 #define AUTHORS "Michael Meskes"
 
 static struct option const long_options[] = {
-  {"link", no_argument, 0, 'l'}, /* deprecated.  FIXME: remove in 2003 */
-  {"dereference", no_argument, 0, 'L'},
-  {"format", required_argument, 0, 'c'},
-  {"filesystem", no_argument, 0, 'f'},
-  {"terse", no_argument, 0, 't'},
+  {"link", no_argument, NULL, 'l'}, /* deprecated.  FIXME: remove in 2003 */
+  {"dereference", no_argument, NULL, 'L'},
+  {"file-system", no_argument, NULL, 'f'},
+  {"filesystem", no_argument, NULL, 'f'}, /* obsolete and undocumented alias */
+  {"format", required_argument, NULL, 'c'},
+  {"terse", no_argument, NULL, 't'},
   {GETOPT_HELP_OPTION_DECL},
   {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
 };
 
-/* Nonzero means we should exit with EXIT_FAILURE upon completion.  */
-static int G_fail;
-
 char *program_name;
 
 /* Return the type of the specified file system.
    Some systems have statfvs.f_basetype[FSTYPSZ]. (AIX, HP-UX, and Solaris)
    Others have statfs.f_fstypename[MFSNAMELEN]. (NetBSD 1.5.2)
    Still others have neither and have to get by with f_type (Linux).  */
-static char *
+static char const *
 human_fstype (STRUCT_STATVFS const *statfsbuf)
 {
 #ifdef STATXFS_FILE_SYSTEM_TYPE_MEMBER_NAME
-  /* Cast away the `const' attribute.  */
-  return (char *) statfsbuf->STATXFS_FILE_SYSTEM_TYPE_MEMBER_NAME;
+  return statfsbuf->STATXFS_FILE_SYSTEM_TYPE_MEMBER_NAME;
 #else
-  char const *type;
   switch (statfsbuf->f_type)
     {
 # if defined __linux__
@@ -127,192 +134,140 @@ human_fstype (STRUCT_STATVFS const *statfsbuf)
 	 combined to produce the #define directives in fs.h.  */
 
     case S_MAGIC_AFFS: /* 0xADFF */
-      type = "affs";
-      break;
+      return "affs";
     case S_MAGIC_DEVPTS: /* 0x1CD1 */
-      type = "devpts";
-      break;
+      return "devpts";
     case S_MAGIC_EXT: /* 0x137D */
-      type = "ext";
-      break;
+      return "ext";
     case S_MAGIC_EXT2_OLD: /* 0xEF51 */
-      type = "ext2";
-      break;
+      return "ext2";
     case S_MAGIC_EXT2: /* 0xEF53 */
-      type = "ext2/ext3";
-      break;
+      return "ext2/ext3";
+    case S_MAGIC_JFS: /* 0x3153464a */
+      return "jfs";
+    case S_MAGIC_XFS: /* 0x58465342 */
+      return "xfs";
     case S_MAGIC_HPFS: /* 0xF995E849 */
-      type = "hpfs";
-      break;
+      return "hpfs";
     case S_MAGIC_ISOFS: /* 0x9660 */
-      type = "isofs";
-      break;
+      return "isofs";
     case S_MAGIC_ISOFS_WIN: /* 0x4000 */
-      type = "isofs";
-      break;
+      return "isofs";
     case S_MAGIC_ISOFS_R_WIN: /* 0x4004 */
-      type = "isofs";
-      break;
+      return "isofs";
     case S_MAGIC_MINIX: /* 0x137F */
-      type = "minix";
-      break;
+      return "minix";
     case S_MAGIC_MINIX_30: /* 0x138F */
-      type = "minix (30 char.)";
-      break;
+      return "minix (30 char.)";
     case S_MAGIC_MINIX_V2: /* 0x2468 */
-      type = "minix v2";
-      break;
+      return "minix v2";
     case S_MAGIC_MINIX_V2_30: /* 0x2478 */
-      type = "minix v2 (30 char.)";
-      break;
+      return "minix v2 (30 char.)";
     case S_MAGIC_MSDOS: /* 0x4d44 */
-      type = "msdos";
-      break;
+      return "msdos";
     case S_MAGIC_FAT: /* 0x4006 */
-      type = "fat";
-      break;
+      return "fat";
     case S_MAGIC_NCP: /* 0x564c */
-      type = "novell";
-      break;
+      return "novell";
     case S_MAGIC_NFS: /* 0x6969 */
-      type = "nfs";
-      break;
+      return "nfs";
     case S_MAGIC_PROC: /* 0x9fa0 */
-      type = "proc";
-      break;
+      return "proc";
     case S_MAGIC_SMB: /* 0x517B */
-      type = "smb";
-      break;
+      return "smb";
     case S_MAGIC_XENIX: /* 0x012FF7B4 */
-      type = "xenix";
-      break;
+      return "xenix";
     case S_MAGIC_SYSV4: /* 0x012FF7B5 */
-      type = "sysv4";
-      break;
+      return "sysv4";
     case S_MAGIC_SYSV2: /* 0x012FF7B6 */
-      type = "sysv2";
-      break;
+      return "sysv2";
     case S_MAGIC_COH: /* 0x012FF7B7 */
-      type = "coh";
-      break;
+      return "coh";
     case S_MAGIC_UFS: /* 0x00011954 */
-      type = "ufs";
-      break;
+      return "ufs";
     case S_MAGIC_XIAFS: /* 0x012FD16D */
-      type = "xia";
-      break;
+      return "xia";
     case S_MAGIC_NTFS: /* 0x5346544e */
-      type = "ntfs";
-      break;
+      return "ntfs";
     case S_MAGIC_TMPFS: /* 0x1021994 */
-      type = "tmpfs";
-      break;
+      return "tmpfs";
     case S_MAGIC_REISERFS: /* 0x52654973 */
-      type = "reiserfs";
-      break;
+      return "reiserfs";
     case S_MAGIC_CRAMFS: /* 0x28cd3d45 */
-      type = "cramfs";
-      break;
+      return "cramfs";
     case S_MAGIC_ROMFS: /* 0x7275 */
-      type = "romfs";
-      break;
+      return "romfs";
+    case S_MAGIC_RAMFS: /* 0x858458f6 */
+      return "ramfs";
+    case S_MAGIC_SQUASHFS: /* 0x73717368 */
+      return "squashfs";
+    case S_MAGIC_SYSFS: /* 0x62656572 */
+      return "sysfs";
 # elif __GNU__
     case FSTYPE_UFS:
-      type = "ufs";
-      break;
+      return "ufs";
     case FSTYPE_NFS:
-      type = "nfs";
-      break;
+      return "nfs";
     case FSTYPE_GFS:
-      type = "gfs";
-      break;
+      return "gfs";
     case FSTYPE_LFS:
-      type = "lfs";
-      break;
+      return "lfs";
     case FSTYPE_SYSV:
-      type = "sysv";
-      break;
+      return "sysv";
     case FSTYPE_FTP:
-      type = "ftp";
-      break;
+      return "ftp";
     case FSTYPE_TAR:
-      type = "tar";
-      break;
+      return "tar";
     case FSTYPE_AR:
-      type = "ar";
-      break;
+      return "ar";
     case FSTYPE_CPIO:
-      type = "cpio";
-      break;
+      return "cpio";
     case FSTYPE_MSLOSS:
-      type = "msloss";
-      break;
+      return "msloss";
     case FSTYPE_CPM:
-      type = "cpm";
-      break;
+      return "cpm";
     case FSTYPE_HFS:
-      type = "hfs";
-      break;
+      return "hfs";
     case FSTYPE_DTFS:
-      type = "dtfs";
-      break;
+      return "dtfs";
     case FSTYPE_GRFS:
-      type = "grfs";
-      break;
+      return "grfs";
     case FSTYPE_TERM:
-      type = "term";
-      break;
+      return "term";
     case FSTYPE_DEV:
-      type = "dev";
-      break;
+      return "dev";
     case FSTYPE_PROC:
-      type = "proc";
-      break;
+      return "proc";
     case FSTYPE_IFSOCK:
-      type = "ifsock";
-      break;
+      return "ifsock";
     case FSTYPE_AFS:
-      type = "afs";
-      break;
+      return "afs";
     case FSTYPE_DFS:
-      type = "dfs";
-      break;
+      return "dfs";
     case FSTYPE_PROC9:
-      type = "proc9";
-      break;
+      return "proc9";
     case FSTYPE_SOCKET:
-      type = "socket";
-      break;
+      return "socket";
     case FSTYPE_MISC:
-      type = "misc";
-      break;
+      return "misc";
     case FSTYPE_EXT2FS:
-      type = "ext2/ext3";
-      break;
+      return "ext2/ext3";
     case FSTYPE_HTTP:
-      type = "http";
-      break;
+      return "http";
     case FSTYPE_MEMFS:
-      type = "memfs";
-      break;
+      return "memfs";
     case FSTYPE_ISO9660:
-      type = "iso9660";
-      break;
+      return "iso9660";
 # endif
     default:
-      type = NULL;
-      break;
+      {
+	unsigned long int type = statfsbuf->f_type;
+	static char buf[sizeof "UNKNOWN (0x%lx)" - 3
+			+ (sizeof type * CHAR_BIT + 3) / 4];
+	sprintf (buf, "UNKNOWN (0x%lx)", type);
+	return buf;
+      }
     }
-
-  if (type)
-    return (char *) type;
-
-  {
-    static char buf[sizeof "UNKNOWN (0x%x)" - 2
-		    + 2 * sizeof (statfsbuf->f_type)];
-    sprintf (buf, "UNKNOWN (0x%x)", statfsbuf->f_type);
-    return buf;
-  }
 #endif
 }
 
@@ -326,22 +281,37 @@ human_access (struct stat const *statbuf)
 }
 
 static char *
-human_time (time_t const *t)
+human_time (struct timespec t)
 {
-  static char str[80];
-  struct tm *tm = localtime (t);
+  static char str[MAX (INT_BUFSIZE_BOUND (intmax_t),
+		       (INT_STRLEN_BOUND (int) /* YYYY */
+			+ 1 /* because YYYY might equal INT_MAX + 1900 */
+			+ sizeof "-MM-DD HH:MM:SS.NNNNNNNNN +ZZZZ"))];
+  struct tm const *tm = localtime (&t.tv_sec);
   if (tm == NULL)
-    {
-      G_fail = 1;
-      return (char *) _("*** invalid date/time ***");
-    }
-  nstrftime (str, sizeof str, "%Y-%m-%d %H:%M:%S.%N %z", tm, 0, 0);
+    return (TYPE_SIGNED (time_t)
+	    ? imaxtostr (t.tv_sec, str)
+	    : umaxtostr (t.tv_sec, str));
+  nstrftime (str, sizeof str, "%Y-%m-%d %H:%M:%S.%N %z", tm, 0, t.tv_nsec);
   return str;
 }
 
+/* Like strcat, but don't return anything and do check that
+   DEST_BUFSIZE is at least a long as strlen (DEST) + strlen (SRC) + 1.
+   The signature is deliberately different from that of strncat.  */
+static void
+xstrcat (char *dest, size_t dest_bufsize, char const *src)
+{
+  size_t dest_len = strlen (dest);
+  size_t src_len = strlen (src);
+  if (dest_bufsize < dest_len + src_len + 1)
+    abort ();
+  memcpy (dest + dest_len, src, src_len + 1);
+}
+
 /* print statfs info */
 static void
-print_statfs (char *pformat, char m, char const *filename,
+print_statfs (char *pformat, size_t buf_len, char m, char const *filename,
 	      void const *data)
 {
   STRUCT_STATVFS const *statfsbuf = data;
@@ -349,64 +319,74 @@ print_statfs (char *pformat, char m, char const *filename,
   switch (m)
     {
     case 'n':
-      strcat (pformat, "s");
+      xstrcat (pformat, buf_len, "s");
       printf (pformat, filename);
       break;
 
     case 'i':
 #if HAVE_STRUCT_STATXFS_F_FSID___VAL
-      strcat (pformat, "x %-8x");
+      xstrcat (pformat, buf_len, "x %-8x");
       printf (pformat, statfsbuf->f_fsid.__val[0], /* u_long */
 	      statfsbuf->f_fsid.__val[1]);
 #else
-      strcat (pformat, "Lx");
+      xstrcat (pformat, buf_len, "Lx");
       printf (pformat, statfsbuf->f_fsid);
 #endif
       break;
 
     case 'l':
-      strcat (pformat, NAMEMAX_FORMAT);
+      xstrcat (pformat, buf_len, NAMEMAX_FORMAT);
       printf (pformat, SB_F_NAMEMAX (statfsbuf));
       break;
     case 't':
 #if HAVE_STRUCT_STATXFS_F_TYPE
-      strcat (pformat, "lx");
-      printf (pformat, (long int) (statfsbuf->f_type));  /* no equiv. */
+      xstrcat (pformat, buf_len, "lx");
+      printf (pformat,
+	      (unsigned long int) (statfsbuf->f_type));  /* no equiv. */
 #else
       fputc ('*', stdout);
 #endif
       break;
     case 'T':
-      strcat (pformat, "s");
+      xstrcat (pformat, buf_len, "s");
       printf (pformat, human_fstype (statfsbuf));
       break;
     case 'b':
-      strcat (pformat, PRIdMAX);
+      xstrcat (pformat, buf_len, PRIdMAX);
       printf (pformat, (intmax_t) (statfsbuf->f_blocks));
       break;
     case 'f':
-      strcat (pformat, PRIdMAX);
+      xstrcat (pformat, buf_len, PRIdMAX);
       printf (pformat, (intmax_t) (statfsbuf->f_bfree));
       break;
     case 'a':
-      strcat (pformat, PRIdMAX);
+      xstrcat (pformat, buf_len, PRIdMAX);
       printf (pformat, (intmax_t) (statfsbuf->f_bavail));
       break;
     case 's':
-      strcat (pformat, "ld");
-      printf (pformat, (long int) (statfsbuf->f_bsize));
+      xstrcat (pformat, buf_len, "lu");
+      printf (pformat, (unsigned long int) (statfsbuf->f_bsize));
+      break;
+    case 'S':
+      {
+	unsigned long int frsize = STATFS_FRSIZE (statfsbuf);
+	if (! frsize)
+	  frsize = statfsbuf->f_bsize;
+	xstrcat (pformat, buf_len, "lu");
+	printf (pformat, frsize);
+      }
       break;
     case 'c':
-      strcat (pformat, PRIdMAX);
+      xstrcat (pformat, buf_len, PRIdMAX);
       printf (pformat, (intmax_t) (statfsbuf->f_files));
       break;
     case 'd':
-      strcat (pformat, PRIdMAX);
+      xstrcat (pformat, buf_len, PRIdMAX);
       printf (pformat, (intmax_t) (statfsbuf->f_ffree));
       break;
 
     default:
-      strcat (pformat, "c");
+      xstrcat (pformat, buf_len, "c");
       printf (pformat, m);
       break;
     }
@@ -414,7 +394,8 @@ print_statfs (char *pformat, char m, char const *filename,
 
 /* print stat info */
 static void
-print_stat (char *pformat, char m, char const *filename, void const *data)
+print_stat (char *pformat, size_t buf_len, char m,
+	    char const *filename, void const *data)
 {
   struct stat *statbuf = (struct stat *) data;
   struct passwd *pw_ent;
@@ -423,14 +404,14 @@ print_stat (char *pformat, char m, char const *filename, void const *data)
   switch (m)
     {
     case 'n':
-      strcat (pformat, "s");
+      xstrcat (pformat, buf_len, "s");
       printf (pformat, filename);
       break;
     case 'N':
-      strcat (pformat, "s");
+      xstrcat (pformat, buf_len, "s");
       if (S_ISLNK (statbuf->st_mode))
 	{
-	  char *linkname = xreadlink (filename);
+	  char *linkname = xreadlink (filename, statbuf->st_size);
 	  if (linkname == NULL)
 	    {
 	      error (0, errno, _("cannot read symbolic link %s"),
@@ -448,107 +429,108 @@ print_stat (char *pformat, char m, char const *filename, void const *data)
 	}
       break;
     case 'd':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_dev);
+      xstrcat (pformat, buf_len, PRIuMAX);
+      printf (pformat, (uintmax_t) statbuf->st_dev);
       break;
     case 'D':
-      strcat (pformat, "x");
-      printf (pformat, (int) statbuf->st_dev);
+      xstrcat (pformat, buf_len, PRIxMAX);
+      printf (pformat, (uintmax_t) statbuf->st_dev);
       break;
     case 'i':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_ino);
+      xstrcat (pformat, buf_len, PRIuMAX);
+      printf (pformat, (uintmax_t) statbuf->st_ino);
       break;
     case 'a':
-      strcat (pformat, "o");
-      printf (pformat, statbuf->st_mode & 07777);
+      xstrcat (pformat, buf_len, "lo");
+      printf (pformat,
+	      (unsigned long int) (statbuf->st_mode & CHMOD_MODE_BITS));
       break;
     case 'A':
-      strcat (pformat, "s");
+      xstrcat (pformat, buf_len, "s");
       printf (pformat, human_access (statbuf));
       break;
     case 'f':
-      strcat (pformat, "x");
-      printf (pformat, statbuf->st_mode);
+      xstrcat (pformat, buf_len, "lx");
+      printf (pformat, (unsigned long int) statbuf->st_mode);
       break;
     case 'F':
-      strcat (pformat, "s");
+      xstrcat (pformat, buf_len, "s");
       printf (pformat, file_type (statbuf));
       break;
     case 'h':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_nlink);
+      xstrcat (pformat, buf_len, "lu");
+      printf (pformat, (unsigned long int) statbuf->st_nlink);
       break;
     case 'u':
-      strcat (pformat, "d");
-      printf (pformat, statbuf->st_uid);
+      xstrcat (pformat, buf_len, "lu");
+      printf (pformat, (unsigned long int) statbuf->st_uid);
       break;
     case 'U':
-      strcat (pformat, "s");
+      xstrcat (pformat, buf_len, "s");
       setpwent ();
       pw_ent = getpwuid (statbuf->st_uid);
       printf (pformat, (pw_ent != 0L) ? pw_ent->pw_name : "UNKNOWN");
       break;
     case 'g':
-      strcat (pformat, "d");
-      printf (pformat, statbuf->st_gid);
+      xstrcat (pformat, buf_len, "lu");
+      printf (pformat, (unsigned long int) statbuf->st_gid);
       break;
     case 'G':
-      strcat (pformat, "s");
+      xstrcat (pformat, buf_len, "s");
       setgrent ();
       gw_ent = getgrgid (statbuf->st_gid);
       printf (pformat, (gw_ent != 0L) ? gw_ent->gr_name : "UNKNOWN");
       break;
     case 't':
-      strcat (pformat, "x");
-      printf (pformat, major (statbuf->st_rdev));
+      xstrcat (pformat, buf_len, "lx");
+      printf (pformat, (unsigned long int) major (statbuf->st_rdev));
       break;
     case 'T':
-      strcat (pformat, "x");
-      printf (pformat, minor (statbuf->st_rdev));
+      xstrcat (pformat, buf_len, "lx");
+      printf (pformat, (unsigned long int) minor (statbuf->st_rdev));
       break;
     case 's':
-      strcat (pformat, PRIuMAX);
+      xstrcat (pformat, buf_len, PRIuMAX);
       printf (pformat, (uintmax_t) (statbuf->st_size));
       break;
     case 'B':
-      strcat (pformat, "u");
-      printf (pformat, (unsigned int) ST_NBLOCKSIZE);
+      xstrcat (pformat, buf_len, "lu");
+      printf (pformat, (unsigned long int) ST_NBLOCKSIZE);
       break;
     case 'b':
-      strcat (pformat, "u");
-      printf (pformat, (unsigned int) ST_NBLOCKS (*statbuf));
+      xstrcat (pformat, buf_len, PRIuMAX);
+      printf (pformat, (uintmax_t) ST_NBLOCKS (*statbuf));
       break;
     case 'o':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_blksize);
+      xstrcat (pformat, buf_len, "lu");
+      printf (pformat, (unsigned long int) statbuf->st_blksize);
       break;
     case 'x':
-      strcat (pformat, "s");
-      printf (pformat, human_time (&(statbuf->st_atime)));
+      xstrcat (pformat, buf_len, "s");
+      printf (pformat, human_time (get_stat_atime (statbuf)));
       break;
     case 'X':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_atime);
+      xstrcat (pformat, buf_len, TYPE_SIGNED (time_t) ? "ld" : "lu");
+      printf (pformat, (unsigned long int) statbuf->st_atime);
       break;
     case 'y':
-      strcat (pformat, "s");
-      printf (pformat, human_time (&(statbuf->st_mtime)));
+      xstrcat (pformat, buf_len, "s");
+      printf (pformat, human_time (get_stat_mtime (statbuf)));
       break;
     case 'Y':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_mtime);
+      xstrcat (pformat, buf_len, TYPE_SIGNED (time_t) ? "ld" : "lu");
+      printf (pformat, (unsigned long int) statbuf->st_mtime);
       break;
     case 'z':
-      strcat (pformat, "s");
-      printf (pformat, human_time (&(statbuf->st_ctime)));
+      xstrcat (pformat, buf_len, "s");
+      printf (pformat, human_time (get_stat_ctime (statbuf)));
       break;
     case 'Z':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_ctime);
+      xstrcat (pformat, buf_len, TYPE_SIGNED (time_t) ? "ld" : "lu");
+      printf (pformat, (unsigned long int) statbuf->st_ctime);
       break;
     default:
-      strcat (pformat, "c");
+      xstrcat (pformat, buf_len, "c");
       printf (pformat, m);
       break;
     }
@@ -556,7 +538,7 @@ print_stat (char *pformat, char m, char const *filename, void const *data)
 
 static void
 print_it (char const *masterformat, char const *filename,
-	  void (*print_func) (char *, char, char const *, void const *),
+	  void (*print_func) (char *, size_t, char, char const *, void const *),
 	  void const *data)
 {
   char *b;
@@ -564,8 +546,10 @@ print_it (char const *masterformat, char const *filename,
   /* create a working copy of the format string */
   char *format = xstrdup (masterformat);
 
-  char *dest = xmalloc (strlen (format) + 1);
-
+  /* Add 2 to accommodate our conversion of the stat `%s' format string
+     to the printf `%llu' one.  */
+  size_t n_alloc = strlen (format) + 2 + 1;
+  char *dest = xmalloc (n_alloc);
 
   b = format;
   while (b)
@@ -583,17 +567,19 @@ print_it (char const *masterformat, char const *filename,
 	  dest[1 + len] = 0;
 	  p += len;
 
+	  b = p + 1;
 	  switch (*p)
 	    {
 	    case '\0':
+	      b = NULL;
+	      /* fall through */
 	    case '%':
 	      putchar ('%');
 	      break;
 	    default:
-	      print_func (dest, *p, filename, data);
+	      print_func (dest, n_alloc, *p, filename, data);
 	      break;
 	    }
-	  b = p + 1;
 	}
       else
 	{
@@ -603,63 +589,60 @@ print_it (char const *masterformat, char const *filename,
     }
   free (format);
   free (dest);
-  fputc ('\n', stdout);
 }
 
-/* stat the filesystem and print what we find */
-static void
-do_statfs (char const *filename, int terse, char const *format)
+/* Stat the file system and print what we find.  */
+static bool
+do_statfs (char const *filename, bool terse, char const *format)
 {
   STRUCT_STATVFS statfsbuf;
-  int i = statfs (filename, &statfsbuf);
 
-  if (i == -1)
+  if (STATFS (filename, &statfsbuf) != 0)
     {
       error (0, errno, _("cannot read file system information for %s"),
 	     quote (filename));
-      return;
+      return false;
     }
 
   if (format == NULL)
     {
       format = (terse
-		? "%n %i %l %t %b %f %a %s %c %d"
+		? "%n %i %l %t %s %S %b %f %a %c %d\n"
 		: "  File: \"%n\"\n"
 		"    ID: %-8i Namelen: %-7l Type: %T\n"
-		"Blocks: Total: %-10b Free: %-10f Available: %-10a Size: %s\n"
-		"Inodes: Total: %-10c Free: %-10d");
+		"Block size: %-10s Fundamental block size: %S\n"
+		"Blocks: Total: %-10b Free: %-10f Available: %a\n"
+		"Inodes: Total: %-10c Free: %d\n");
     }
 
   print_it (format, filename, print_statfs, &statfsbuf);
+  return true;
 }
 
 /* stat the file and print what we find */
-static void
-do_stat (char const *filename, int follow_links, int terse,
+static bool
+do_stat (char const *filename, bool follow_links, bool terse,
 	 char const *format)
 {
   struct stat statbuf;
-  int i = ((follow_links == 1)
-	   ? stat (filename, &statbuf)
-	   : lstat (filename, &statbuf));
 
-  if (i == -1)
+  if ((follow_links ? stat : lstat) (filename, &statbuf) != 0)
     {
       error (0, errno, _("cannot stat %s"), quote (filename));
-      return;
+      return false;
     }
 
   if (format == NULL)
     {
-      if (terse != 0)
+      if (terse)
 	{
-	  format = "%n %s %b %f %u %g %D %i %h %t %T %X %Y %Z %o";
+	  format = "%n %s %b %f %u %g %D %i %h %t %T %X %Y %Z %o\n";
 	}
       else
 	{
-	  /* tmp hack to match orignal output until conditional implemented */
-	  i = statbuf.st_mode & S_IFMT;
-	  if (i == S_IFCHR || i == S_IFBLK)
+	  /* Temporary hack to match original output until conditional
+	     implemented.  */
+	  if (S_ISBLK (statbuf.st_mode) || S_ISCHR (statbuf.st_mode))
 	    {
 	      format =
 		"  File: %N\n"
@@ -674,28 +657,29 @@ do_stat (char const *filename, int follow_links, int terse,
 	      format =
 		"  File: %N\n"
 		"  Size: %-10s\tBlocks: %-10b IO Block: %-6o %F\n"
-		"Device: %Dh/%dd\tInode: %-10i  Links: %-5h\n"
+		"Device: %Dh/%dd\tInode: %-10i  Links: %h\n"
 		"Access: (%04a/%10.10A)  Uid: (%5u/%8U)   Gid: (%5g/%8G)\n"
 		"Access: %x\n" "Modify: %y\n" "Change: %z\n";
 	    }
 	}
     }
   print_it (format, filename, print_stat, &statbuf);
+  return true;
 }
 
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
     {
       printf (_("Usage: %s [OPTION] FILE...\n"), program_name);
       fputs (_("\
-Display file or filesystem status.\n\
+Display file or file system status.\n\
 \n\
-  -f, --filesystem      display filesystem status instead of file status\n\
+  -f, --file-system     display file system status instead of file status\n\
   -c  --format=FORMAT   use the specified FORMAT instead of the default\n\
   -L, --dereference     follow links\n\
   -t, --terse           print the information in terse form\n\
@@ -704,40 +688,40 @@ Display file or filesystem status.\n\
       fputs (VERSION_OPTION_DESCRIPTION, stdout);
 
       fputs (_("\n\
-The valid format sequences for files (without --filesystem):\n\
+The valid format sequences for files (without --file-system):\n\
 \n\
-  %A   Access rights in human readable form\n\
   %a   Access rights in octal\n\
-  %B   The size in bytes of each block reported by `%b'\n\
+  %A   Access rights in human readable form\n\
   %b   Number of blocks allocated (see %B)\n\
+  %B   The size in bytes of each block reported by %b\n\
 "), stdout);
       fputs (_("\
-  %D   Device number in hex\n\
   %d   Device number in decimal\n\
-  %F   File type\n\
+  %D   Device number in hex\n\
   %f   Raw mode in hex\n\
-  %G   Group name of owner\n\
+  %F   File type\n\
   %g   Group ID of owner\n\
+  %G   Group name of owner\n\
 "), stdout);
       fputs (_("\
   %h   Number of hard links\n\
   %i   Inode number\n\
-  %N   Quoted File name with dereference if symbolic link\n\
   %n   File name\n\
-  %o   IO block size\n\
+  %N   Quoted file name with dereference if symbolic link\n\
+  %o   I/O block size\n\
   %s   Total size, in bytes\n\
-  %T   Minor device type in hex\n\
   %t   Major device type in hex\n\
+  %T   Minor device type in hex\n\
 "), stdout);
       fputs (_("\
-  %U   User name of owner\n\
   %u   User ID of owner\n\
-  %X   Time of last access as seconds since Epoch\n\
+  %U   User name of owner\n\
   %x   Time of last access\n\
-  %Y   Time of last modification as seconds since Epoch\n\
+  %X   Time of last access as seconds since Epoch\n\
   %y   Time of last modification\n\
-  %Z   Time of last change as seconds since Epoch\n\
+  %Y   Time of last modification as seconds since Epoch\n\
   %z   Time of last change\n\
+  %Z   Time of last change as seconds since Epoch\n\
 \n\
 "), stdout);
 
@@ -751,13 +735,15 @@ Valid format sequences for file systems:\n\
   %f   Free blocks in file system\n\
 "), stdout);
       fputs (_("\
-  %i   File System id in hex\n\
+  %i   File System ID in hex\n\
   %l   Maximum length of filenames\n\
   %n   File name\n\
-  %s   Optimal transfer block size\n\
-  %T   Type in human readable form\n\
+  %s   Block size (for faster transfers)\n\
+  %S   Fundamental block size (for block counts)\n\
   %t   Type in hex\n\
+  %T   Type in human readable form\n\
 "), stdout);
+      printf (USAGE_BUILTIN_WARNING, PROGRAM_NAME);
       printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
   exit (status);
@@ -768,10 +754,11 @@ main (int argc, char *argv[])
 {
   int c;
   int i;
-  int follow_links = 0;
-  int fs = 0;
-  int terse = 0;
+  bool follow_links = false;
+  bool fs = false;
+  bool terse = false;
   char *format = NULL;
+  bool ok = true;
 
   initialize_main (&argc, &argv);
   program_name = argv[0];
@@ -788,15 +775,20 @@ main (int argc, char *argv[])
 	case 'c':
 	  format = optarg;
 	  break;
+
 	case 'l': /* deprecated */
+	  error (0, 0, _("Warning: `-l' is deprecated; use `-L' instead"));
+	  /* fall through */
 	case 'L':
-	  follow_links = 1;
+	  follow_links = true;
 	  break;
+
 	case 'f':
-	  fs = 1;
+	  fs = true;
 	  break;
+
 	case 't':
-	  terse = 1;
+	  terse = true;
 	  break;
 
 	case_GETOPT_HELP_CHAR;
@@ -810,17 +802,14 @@ main (int argc, char *argv[])
 
   if (argc == optind)
     {
-      error (0, 0, _("too few arguments"));
+      error (0, 0, _("missing operand"));
       usage (EXIT_FAILURE);
     }
 
   for (i = optind; i < argc; i++)
-    {
-      if (fs == 0)
-	do_stat (argv[i], follow_links, terse, format);
-      else
-	do_statfs (argv[i], terse, format);
-    }
+    ok &= (fs
+	   ? do_statfs (argv[i], terse, format)
+	   : do_stat (argv[i], follow_links, terse, format));
 
-  exit (G_fail ? EXIT_FAILURE : EXIT_SUCCESS);
+  exit (ok ? EXIT_SUCCESS : EXIT_FAILURE);
 }
